// Copyright (c) 2017, teja. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

library jaguar_session_mongo.src;

import 'dart:io';
import 'dart:async';
import 'dart:convert';
import 'dart:collection';
import 'package:mongo_dart/mongo_dart.dart';
import 'package:jaguar/jaguar.dart';
import 'package:jaguar_session/jaguar_session.dart';
import 'package:jaguar_data_store/jaguar_data_store.dart';
import 'package:jaguar_mongo_data_store/jaguar_mongo_data_store.dart';
import 'package:jaguar_serializer/serializer.dart';

part 'model.dart';

/// MongoDb based session manager that stores session identifier on Cookie
///
/// Stores session data on MongoDb. Stores the session identifier on Cookie.
class MgoCookieSession implements SessionManager {
  /// Name of the cookie on which session identifier is stored
  final String cookieName;

  final MongoDataStore<_SessionData> dataStore;

  MgoCookieSession(Db db, [this.cookieName = 'session'])
      : dataStore =
            new MongoDataStore<_SessionData>(_serializer, "session", db);

  /// Incoming session data parsed from the request
  Map<String, String> _inValues = {};

  /// Outgoing session data that must be written to the response
  Map<String, String> _outValues = {};

  /// Should the session be updated?
  bool _shouldUpdate = false;

  String _id;

  UnmodifiableMapView<String, String> get inValues =>
      new UnmodifiableMapView<String, String>(_inValues);

  UnmodifiableMapView<String, String> get outValues =>
      new UnmodifiableMapView<String, String>(_outValues);

  /// Returns a session value by given [key]
  String getInValue(String key) => _inValues[key];

  /// Creates a new session
  Future<Null> createSession(Map<String, String> values) async {
    _shouldUpdate = true;
    await deleteSession();
    {
      _SessionData data = new _SessionData();
      data.data = values;
      _id = await dataStore.insert(data);
    }
    {
      final _SessionData data = await dataStore.getById(_id);
      _inValues = data.data;
      _outValues = new Map<String, String>.from(data.data);
    }
    return null;
  }

  /// Update session info
  Future<Null> updateSession(Map<String, String> values) async {
    if (_id is! String) {
      throw new Exception("Doesn't have a session!");
    }
    _outValues.addAll(values);
    _shouldUpdate = true;
  }

  /// Deletes the session
  Future<Null> deleteSession() async {
    String myId = _id;
    _id = null;
    _inValues.clear();
    _outValues.clear();
    _shouldUpdate = true;

    if (myId is String) {
      await dataStore.deleteById(myId);
    }
  }

  /// Parses session from the given [request]
  Future<Null> parseRequest(Request request) async {
    _id = null;
    _inValues = {};
    _outValues = {};
    for (Cookie cook in request.cookies) {
      if (cook.name == cookieName) {
        final String id = _decode(cook.value);
        if (_isValidMgoId(id)) {
          _id = id;
          final _SessionData data = await dataStore.getById(_id);
          _inValues = data.data;
          _outValues = new Map<String, String>.from(data.data);
        }
        break;
      }
    }
  }

  /// Updates [response] with session details
  Future<Response> writeResponse(Request request, Response resp) async {
    if (_shouldUpdate) {
      if (_id is String) {
        _SessionData data = new _SessionData();
        data.id = _id;
        data.data = _outValues;
        await dataStore.updateById(_id, data);
        Cookie cook = new Cookie(cookieName, _encode(_id));
        resp.cookies.add(cook);
      } else {
        Cookie cook = new Cookie(cookieName, _encode('0' * 24));
        resp.cookies.add(cook);
      }
    }
    return resp;
  }

  /// Validates the current session
  ///
  /// Must throw exception on validation failure
  Future<Null> validate() async {}

  String _encode(String myId) {
    String str = myId;
    //TODO encrypt
    return const Base64Codec.urlSafe().encode(str.codeUnits);
  }

  String _decode(String data) {
    String str =
        new String.fromCharCodes(const Base64Codec.urlSafe().decode(data));
    //TODO decrypt
    return str;
  }
}

bool _isValidMgoId(String hexString) {
  if (hexString is! String) return false;

  if (hexString.length != 24) return false;

  if (hexString == ('0' * 24)) return false;

  if (!hexString.toUpperCase().codeUnits.every(_isCharHex)) return false;

  return true;
}

bool _isCharHex(int code) {
  if (code >= 48 && code <= 57) return true;

  if (code >= 65 && code <= 70) return true;

  return false;
}
